import ConfigParser
import trac
import time
import trac_fedmsg_plugin
import requests
from tracrpc import IXMLRPCHandler, expose_rpc
from trac.ticket.model import Ticket
import os

if 'get_request_object' in dir(trac_fedmsg_plugin):
    from trac_fedmsg_plugin import get_request_object
else:
    import inspect
    def get_request_object():
        for frame in (f[0] for f in inspect.stack()):
            if 'req' in frame.f_locals:
                return frame.f_locals['req']


def recursive_str(dct):
    """ This function makes sure that dct is json serializable. """
    if not hasattr(dct, '__iter__'):
            return str(dct)
    for key in dct:
        if isinstance(dct[key], dict):
            dct[key] = recursive_str(dct[key])
        elif isinstance(dct[key], list):
            dct[key] = map(recursive_str, dct[key])
        elif isinstance(dct[key], unicode):
            dct[key] = str(dct[key].encode('utf-8'))
        else:
            # If it's not a dict or list, just run an str() over it
            dct[key] = str(dct[key])
    return dct


def attachment2dict(attachment):
    attrs = ['parent_realm', 'parent_id', 'filename', 'description', 'size',
             'date', 'author']
    return dict([(attr, getattr(attachment, attr)) for attr in attrs])


class BassetPlugin(trac_fedmsg_plugin.FedmsgPlugin):
    banned_fields = []
    trac.core.implements(
    	trac.ticket.api.ITicketChangeListener,
        trac.wiki.api.IWikiChangeListener,
        trac.attachment.IAttachmentChangeListener,
        trac.perm.IPermissionPolicy,
        trac.perm.IPermissionRequestor,
        trac.web.api.IRequestFilter,
        IXMLRPCHandler,
    ) 

    def __init__(self, *args, **kwargs):
        # We explicitly do not initialize fedmsg
        trac.core.Component.__init__(self, *args, **kwargs)

        self.config = ConfigParser.SafeConfigParser()
        self.config.read(['trac-baseet.default.cfg',
                          '/etc/basset/trac-basset.default.cfg',
                          'trac-basset.cfg',
                          '/etc/basset/trac-basset.cfg'])

    # XMLRPC stuff
    def get_permission_actions(self):
        return ['ANTISPAM']

    def xmlrpc_namespace(self):
        return 'basset'

    def xmlrpc_methods(self):
        yield ('ANTISPAM', ((bool, str),), self.blockUser)
        yield ('ANTISPAM', ((bool, int, int),), self.deleteTicketChange)

    # Stuff related to blocking
    def _get_block_filename(self, username):
        return os.path.join(
            self.config.get('block', 'directory'),
            username)

    def blockUser(self, req, username):
        """ Block a user from this trac instance. """
        try:
            open(self._get_block_filename(username), 'w').close()
            return True
        except:
            return False

    def check_permission(self, action, username, resource, perm):
        if os.path.exists(self._get_block_filename(username)):
            return False

    def pre_process_request(self, req, handler):
        if self.check_permission(None, req.authname, None, None) is False:
            raise Exception('Denied ' + req.authname)
            # TODO: Perhaps at some point return a nice error message?
            return None
        return handler

    def post_process_request(self, req, template, data, content_type, method=None):
        return (template, data, content_type)

    # Stuff related to undoing a ticket change
    def deleteTicketChange(self, req, ticket_id, changetime):
        """ Undoes a specific ticket change. """
        ticket = Ticket(req, ticket_id)
        ticket.delete_change(changetime, is_ts=True)
        return True

    # Stuff related to submitting
    def publish(self, topic, **msg):
        current_request = get_request_object()
        msg['request_headers'] = dict(get_request_object()._inheaders)
        msg['agent'] = trac_fedmsg_plugin.currently_logged_in_user()
        msg['instance'] = trac_fedmsg_plugin.env2dict(self.env)
        msg = recursive_str(msg)

        basset_request = {'action': 'trac.%s' % topic,
                          'time': int(time.time()),
                          'data': msg}

        url = self.config.get('submit', 'url')
        user = self.config.get('submit', 'username')
        pwd = self.config.get('submit', 'password')

        requests.post(url, auth=(user, pwd), json=basset_request)

    def attachment_added(self, attachment):
        self.publish(topic='attachment.added',
                     attachment=attachment2dict(attachment))

    def attachment_deleted(self, attachment):
        self.publish(topic='attachment.deleted',
                     attachment=attachment2dict(attachment))

    def attachment_reparented(self, attachment, old_parent_realm, old_parent_id):
        self.publish(topic='attachment.reparented', 
                     attachment=attachment2dict(attachment),
                     old_parent_realm=old_parent_realm,
                     old_parent_id=old_parent_id) 
